/*	$NetBSD$	*/

/*
 * Copyright (c) 2005 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Steve C. Woodford for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed for the NetBSD Project by
 *	Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	_MPC8XX_CPU_H_
#define	_MPC8XX_CPU_H_

/* Important kernel virtual addresses */
#define	PPC_KSEG0_BASE		0xc0000000
#define	PPC_KSEG1_BASE		0xe0000000
#define	PPC_KSEG0_SIZE		0x20000000
#define	PPC_KSEG1_SIZE		0x1f000000

#define	PPC_KSEG0_TO_PA(x)	((unsigned)(x) & (PPC_KSEG0_SIZE - 1))
#define	PPC_PA_TO_KSEG0(x)	((unsigned)(x) | PPC_KSEG0_BASE)

#define	MPC_IMMR_KVA		0xff000000

#ifdef _KERNEL
#include <powerpc/mpc8xx/cpuconf.h>

extern char bootpath[];

/*
 * System clocks, derived from EXTCLK/OSCM/MODCK
 */
struct mpc8xx_clocks {
	u_int	pitclk;		/* Programmable interval timer clock */
	u_int	jdbck;		/* DPLL output clock */
	u_int	syncclk;	/* Syncronisation clock */
	u_int	brgclk;		/* Baud rate generator clock */
	struct {
		u_int	gclk;	/* General system clock (e.g. cpu core) */
		u_int	gclk_50;/* Memory controller & external clock */
		u_int	tmbclk;	/* Time base / Decrementer clock */
		u_int	clkout;	/* CLKOUT pin */
	} pwr[2];
#define	MPC8XX_CLK_HIGH	0	/* High power mode clocks (normal) */
#define	MPC8XX_CLK_LOW	1	/* Low power mode clocks */
};
extern struct mpc8xx_clocks mpc8xx_clocks;
extern struct propdb *mpc8xx_propdb;

void mpc8xx_init(void);
void mpc8xx_probe_clocks(u_int, u_int, u_int);
void mpc8xx_startup(const char *);
void mpc8xx_dumpsys(void);
void mpc8xx_install_extint(void (*)(struct intrframe *));
void identifycpu(char *);
#endif /* _KERNEL */

#include <powerpc/cpu.h>

#if defined(_KERNEL) && !defined(_LOCORE)
static __inline int
enable_interrupts(void)
{
	register_t msr;

	msr = mfmsr();
	mtmsr(msr | PSL_EE);
	return (int)(msr & PSL_EE);
}

static __inline int
disable_interrupts(void)
{
	register_t msr;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	__asm __volatile("sync");
	return (int)(msr & PSL_EE);
}

static __inline void
restore_interrupts(int oldee)
{

	mtmsr((mfmsr() & ~PSL_EE) | (register_t)oldee);
	__asm __volatile("sync");
}
#endif	/* defined(_KERNEL) && !defined(_LOCORE) */

#endif	/* _MPC8XX_CPU_H_ */
