/*	$NetBSD$	*/

/*
 * Copyright (c) 2005 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Steve C. Woodford for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed for the NetBSD Project by
 *	Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	_MPC8XX_PMAP_H_
#define	_MPC8XX_PMAP_H_

#include <powerpc/mpc8xx/pte.h>

/* 
 * Extra flags to pass to pmap_enter() -- make sure they don't conflict
 * w/PMAP_CANFAIL or PMAP_WIRED
 */
#define	PME_NOCACHE	0x100
#define	PMAP_NC		PME_NOCACHE	/* XXX: OEA pmap compat. for bus_dma */

#ifndef _LOCORE
/*
 * Pmap stuff
 */
struct pmap {
	int pm_casid;			/* TLB process/translation CASID */
	int pm_casidgen;		/* its generation number */
	int pm_refs;			/* ref count */
	int pm_remove_all;
	pd_entry_t *pm_l1;		/* Pointer to pmap's L1 MMU table */
	struct pmap_statistics pm_stats; /* pmap statistics */
};

typedef	struct pmap *pmap_t;

#ifdef	_KERNEL

#define	PMAP_GROWKERNEL
extern struct pmap kernel_pmap_;
#define	pmap_kernel()			(&kernel_pmap_)

#define	PMAP_MOD	0x1
#define	PMAP_REF	0x2
#define	PMAP_CLEAR	0x100

#define pmap_clear_modify(pg)	 (pmap_check_attr((pg), PMAP_MOD|PMAP_CLEAR))
#define	pmap_clear_reference(pg) (pmap_check_attr((pg), PMAP_REF|PMAP_CLEAR))
#define	pmap_is_modified(pg)	 (pmap_check_attr((pg), PMAP_MOD))
#define	pmap_is_referenced(pg)	 (pmap_check_attr((pg), PMAP_REF))

#define	pmap_phys_address(x)		(x)

#define	pmap_resident_count(pmap)	((pmap)->pm_stats.resident_count)
#define	pmap_wired_count(pmap)		((pmap)->pm_stats.wired_count)

void pmap_unwire(struct pmap *, vaddr_t);
vaddr_t pmap_init_kseg0(vaddr_t, vaddr_t, size_t, paddr_t);
void pmap_bootstrap(paddr_t, psize_t);
boolean_t pmap_extract(struct pmap *, vaddr_t, paddr_t *);
boolean_t pmap_check_attr(struct vm_page *, u_int);
int pmap_write_fault_fixup(struct pmap *, vaddr_t, int);
int pmap_read_fault_fixup(struct pmap *, vaddr_t, int);

#define PMAP_NEED_PROCWR
void pmap_procwr(struct proc *, vaddr_t, size_t);

/*
 * Alternate mapping hooks for pool pages.  Avoids thrashing the TLB.
 *
 * Note: This won't work if we have more memory than can be direct-mapped
 * VA==PA all at once.  But pmap_copy_page() and pmap_zero_page() will have
 * this problem, too.
 */
#define	PMAP_MAP_POOLPAGE(pa)	PPC_PA_TO_KSEG0(pa)
#define	PMAP_UNMAP_POOLPAGE(pa)	PPC_KSEG0_TO_PA(pa)

static __inline paddr_t vtophys(vaddr_t);

static __inline paddr_t
vtophys(vaddr_t va)
{
	paddr_t pa;

	if (pmap_extract(pmap_kernel(), va, &pa))
		return pa;
	return (0);
}
#endif	/* _KERNEL */
#endif	/* _LOCORE */
#endif	/* _MPC8XX_PMAP_H_ */
