/*	$NetBSD$	*/

/*
 * Copyright (c) 2005 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Steve C. Woodford for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed for the NetBSD Project by
 *	Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _MPC_PTE_H
#define	_MPC_PTE_H

/*
 * Note: The pmap uses "Mode 1 Protection Resulution", so only
 * values for Mode 1 are shown below.
 */

/*
 * L1 segment descriptor format
 */
#define	L1_L2BA_MASK	0xfffff000u	/* L2 table base address */
#define	L1_APG(g)	((g)&0xf)<<5)	/* Access protection group (unused) */
#define	L1_G		(1u<<4)		/* Guarded memory */
#define	L1_PS_MASK	0x0c
#define	L1_PS_SMALL	0x00		/* L1 Page size: Small (4/16 KB) */
#define	L1_PS_512KB	0x04		/* L1 Page size: 512KB */
#define	L1_PS_8MB	0x0c		/* L1 Page size: 8MB */
#define	L1_COPY_BACK	0x00		/* Copy-back cache policy */
#define	L1_WRITE_THRU	0x02		/* Write-through cache policy */
#define	L1_VALID	0x01		/* L1 segment entry is valid */

#define	L1_VA2IDX(va)	(((vaddr_t)(va)) >> 22)
#define	L1_PDE2PA(pde)	((paddr_t)((pde) & L1_L2BA_MASK))

/*
 * L2 page descriptor format
 */
#define	L2_RPN_MASK	0xfffff000u	/* Physical (Real) page number */
#define	L2_PP_MASK	0x00000e00	/* Protection mask */
#define	L2_PP_NONE	0x00000200	/* No access */
#define	L2_PP_K_RO	0x00000600	/* Kernel Exec/Read only */
#define	L2_PP_K_RDWR	0x00000000	/* Kernel Exec/Read/Write */
#define	L2_PP_U_RO	0x00000c00	/* User Exec/Read only */
#define	L2_PP_U_RDWR	0x00000800	/* User Exec/Read/Write */
#define	L2_CHANGE	0x00000100	/* Change bit */
#define	L2_SZ_MASK	0x000000f8	/* Page size mask */
#define	L2_SZ_4KB	0x000000f0	/* 4KB page size */
#define	L2_SZ_LARGE	0x000000f8	/* 16KB or larger page size */
#define	L2_SHARED	0x00000004	/* Shared page (ignore ASID) */
#define	L2_CI		0x00000002	/* Cache inhibit */
#define	L2_VALID	0x00000001	/* L2 entry is valid */

#define	L2_VA2IDX(va)	((((vaddr_t)(va)) >> 12) & 0x3ffu)
#define	L2_PTE2PA(pte)	((paddr_t)((pte) & L2_RPN_MASK))
#define	L2_PTES		1024
#define	L2_PTMAP	(PAGE_SIZE * L2_PTES)	/* One L2 can map 4MB memory */

#ifndef _LOCORE
typedef uint32_t pd_entry_t;	/* L1 descriptor table entry */
typedef uint32_t pt_entry_t;	/* L2 page table entry */
#endif

#endif /* _MPC_PTE_H */
