/*	$NetBSD$	*/

/*
 * Copyright (c) 2005 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Steve C. Woodford for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed for the NetBSD Project by
 *	Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This is a version of the locore.S used for the Analogue & Micro
 * Adder875 evaluation board.
 */

#include "opt_ddb.h"
#include "opt_lockdebug.h"
#include "opt_ddbparam.h"
#include "assym.h"

#include <sys/syscall.h>

#include <machine/param.h>
#include <machine/pte.h>
#include <machine/psl.h>
#include <machine/trap.h>

#include <powerpc/spr.h>

#define _NOREGNAMES		/* Allows use of %rN */
#include <machine/asm.h>

#include <powerpc/mpc8xx/mpc885dev.h>

/*
 * Some instructions gas doesn't understand (yet?)
 */
#define	bdneq	bdnzf 2,

#define	INTSTK	8192			/* 8K interrupt stack */

/*
 * Globals
 */
	.data
GLOBAL(esym)
	.long	0			/* end of symbol table */
GLOBAL(proc0paddr)
	.long	0			/* proc0 p_addr */


/*
 * This symbol is here for the benefit of kvm_mkdb, and is supposed to
 * mark the start of kernel text.
 */
	.text
	.globl	_C_LABEL(kernel_text)
_C_LABEL(kernel_text):

	.globl	__start
__start:
	mr	%r31, %r3	/* Save parameter from bootloader */

	/* Ensure translation and interrupts are disabled */
	li	%r7, 0
	mtmsr	%r7
	sync
	isync

#define	MPC_IMMR_PA	0xff000000

	/* Force IMMR to 0xFF000000 */
	lis	%r7, MPC_IMMR_PA@ha
	addi	%r7, %r7, MPC_IMMR_PA@l
	mtspr	SPR_IMMR, %r7

	/* Invalidate all TLBs. */
	lis	%r7, (Mx_CTR_CIDEF | MD_CTR_TWAM)@h
	li	%r0, 0x0
	mtspr	SPR_MD_CTR, %r7
	mtspr	SPR_MI_CTR, %r0
	tlbia

	/*
	 * Zero out the .bss section
	 */
	lis	%r3, _C_LABEL(_edata)-PPC_KSEG0_BASE-4@ha
	addi	%r3, %r3, _C_LABEL(_edata)-PPC_KSEG0_BASE-4@l
	lis	%r8, _C_LABEL(end)-PPC_KSEG0_BASE@ha
	addi	%r8, %r8, _C_LABEL(end)-PPC_KSEG0_BASE@l
	li	%r4, 0x0
1:	stwu	%r4, 4(%r3)
	cmpw	%r3, %r8
	bne	1b

	/*
	 * If RedBoot has passed us a board_info, copy it someplace
	 * where we will be able to access it with the MMU enabled.
	 */
	cmpw	%r4, %r31		/* NULL parameter? */
	beq	2f			/* Yup */

	/* Check alignment */
	andi.	%r4, %r31, 0x3
	bne	2f			/* Misaligned */

	/* Check if the parameter is located in SDRAM (XXX: assume 16MB max) */
	li	%r4, 0x0100
	srwi	%r5, %r31, 16
	cmpw	%r5, %r4
	bge	2f			/* Not in SDRAM */

	/* Check the tag */
	lis	%r4, REDBOOT_BI_TAG@h
	ori	%r4, %r4, REDBOOT_BI_TAG@l
	ldint	%r5, 0(%r31)
	cmpw	%r4, %r5
	bne	2f			/* Invalid tag */

	/*
	 * Let's assume we have a valid board_info.
	 * Copy it to PA 0x2000. init_adder() will copy it somewhere
	 * more permanent after performing some more validity checks.
	 */
#define	BOARD_ID_TEMP_PA	0x2000
	li	%r4, (BOARD_ID_TEMP_PA - 4)
	addi	%r5, %r4, (BOARD_INFO_SZ + 4)
	addi	%r6, %r31, -4
1:	lwzu	%r7, 4(%r6)
	stwu	%r7, 4(%r4)
	cmpw	%r4, %r5
	bne	1b
	lis	%r31, (BOARD_ID_TEMP_PA + PPC_KSEG0_BASE)@ha
	addi	%r31, %r31, (BOARD_ID_TEMP_PA + PPC_KSEG0_BASE)@l
	b	3f
2:	li	%r31, 0			/* Invalid board_info */
3:

	/*
	 * Fix up the TLB to map the kernel into KSEG0
	 * (Make sure to set up a temporary stack)
	 */
	addi	%r1, %r3, INTSTK
	li	%r7, 0xf
	add	%r1, %r1, %r7
	andc	%r1, %r1, %r7		/* SP = roundup(SP, 0x10) */
	lis	%r4, MPC_IMMR_PA@ha
	addi	%r4, %r4, MPC_IMMR_PA@l
	bl	_C_LABEL(pmap_init_tlb)

	/*
	 * Switch to KSEG0
	 */
	li	%r7, (PSL_DR | PSL_IR)
	mtsrr1	%r7
	lis	%r7, .Lmapped_start@ha
	addi	%r7, %r7, .Lmapped_start@l
	mtsrr0	%r7
	sync
	isync
	rfi

.Lmapped_start:
	/*
	 * Ahhh, that's better. Running mapped now.
	 */
	lis	%r5, _C_LABEL(_end)@ha
	addi	%r5, %r5,_C_LABEL(_end)@l
	li	%r7, PGOFSET
	add	%r5, %r5, %r7
	andc	%r5, %r5, %r7		/* _end = roundup(_end, NBPG) */

	/*
	 * Setup cpu_info
	 */
	INIT_CPUINFO(%r5,%r1,%r7,%r8)

#ifdef SYMTAB_SPACE
	lis	%r3, _C_LABEL(db_symtab)@ha
	addi	%r3, %r3, _C_LABEL(db_symtab)@l
	lis	%r4, _C_LABEL(db_symtab)+SYMTAB_SPACE@ha
	addi	%r4, %r4, _C_LABEL(db_symtab)+SYMTAB_SPACE@l
#endif

	/*
	 * Machine Dependent Initialisation
	 *
	 * r3 = ssym (start of symbol table)
	 * r4 = esym (end of symbol table)
	 * r5 = end-of-kernel
	 * r6 = bootloader parameter
	 */
	mr	%r6, %r31
	bl	_C_LABEL(init_adder)

	/*
	 * Invoke the MI kernel startup routing, which should _never_ return
	 */
	bl	_C_LABEL(main)
1:	b	1b


/*
 * Common PowerPC thread/context handlers
 */
#include <powerpc/powerpc/locore_subr.S>

/*
 * Pull in common trap vector code.
 */
#include <powerpc/mpc8xx/trap_subr.S>

/*
 * Include PIO routines
 */
#include <powerpc/powerpc/pio_subr.S>
