/*	$NetBSD$	*/

/*
 * Copyright (c) 2005 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Steve C. Woodford for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed for the NetBSD Project by
 *	Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * NOTICE: This is not a standalone file.  to use it, #include it in
 * your port's locore.S, like so:
 *
 *	#include <powerpc/mpc8xx/trap_subr.S>
 */

/*
 * FRAME_SETUP assumes:
 *	SPRG1		Original SRR0
 *	SPRG2		Original SRR1
 *	SPRG3		Original SP (r1)
 *	savearea	r28-r31, DAR, DSISR
 *	r28		LR
 *	r29		CR
 *	r30		Exception number
 *	r31		Scratch
 *	r1		Kernel stack
 *
 * Translation is already enabled.
 */
#define	FRAME_SETUP(savearea)						    \
	mfsprg3	%r31;				/* Fetch saved SP (r1) */   \
	stregu	%r31, -FRAMELEN(%r1);		/* Save in callframe */	    \
	streg	%r0, (FRAME_0+8)(%r1);		/* Save r0 */		    \
	streg	%r31, (FRAME_1+8)(%r1);		/* Save SP */		    \
	streg	%r2, (FRAME_2+8)(%r1);		/* Save r2 */		    \
	streg	%r28, (FRAME_LR+8)(%r1);	/* Save LR */		    \
	streg	%r29, (FRAME_CR+8)(%r1);	/* Save CR */		    \
	streg	%r30, (FRAME_EXC+8)(%r1);	/* Save Exception number */ \
	GET_CPUINFO(%r2);						    \
	lmw	%r28, savearea(%r2);					    \
	stmw	%r3, (FRAME_3+8)(%r1);		/* Save r3-r31 */	    \
	lmw	%r30, (savearea+16)(%r2);				    \
	mfxer	%r3;							    \
	mfctr	%r4;							    \
	mfsprg1	%r5;				/* Fetch saved SRR0 */	    \
	mfsprg2	%r6;				/* Fetch saved SRR1 */	    \
	streg	%r3, (FRAME_XER+8)(%r1);	/* Save XER */		    \
	streg	%r4, (FRAME_CTR+8)(%r1);	/* Save CTR */		    \
	streg	%r5, (FRAME_SRR0+8)(%r1);	/* Save SRR0 */		    \
	streg	%r6, (FRAME_SRR1+8)(%r1);	/* Save SRR1 */		    \
	streg	%r30, (FRAME_DAR+8)(%r1);	/* Save DAR */		    \
	streg	%r31, (FRAME_DSISR+8)(%r1)	/* Save DSISR */

#define	FRAME_LEAVE							    \
	ldreg	%r2, (FRAME_CTR+8)(%r1);	/* Fetch CTR */		    \
	ldreg	%r3, (FRAME_XER+8)(%r1);	/* Fetch XER */		    \
	ldreg	%r4, (FRAME_LR+8)(%r1);		/* Fetch LR */		    \
	ldreg	%r5, (FRAME_CR+8)(%r1);		/* Fetch CR */		    \
	ldreg	%r6, (FRAME_SRR0+8)(%r1);	/* Fetch SRR0 */	    \
	ldreg	%r7, (FRAME_SRR1+8)(%r1);	/* Fetch SRR1 */	    \
	ldreg	%r8, (FRAME_1+8)(%r1);		/* Fetch SP (r1) */	    \
	mtctr	%r2;				/* Restore CTR */	    \
	mtxer	%r3;				/* Restore XER */	    \
	mtlr	%r4;				/* Restore LR */	    \
	mfmsr	%r0;				/* Exceptions not recov. */ \
	andi.	%r0, %r0, ~(PSL_ME|PSL_RI|PSL_EE)@l;			    \
	mtmsr	%r0;							    \
	isync;								    \
	mtcr	%r5;				/* Restore CR */	    \
	mtsprg1	%r6;				/* Temporarily save SRR0 */ \
	mtsprg2	%r7;				/* Temporarily save SRR1 */ \
	mtsprg3	%r8;				/* Temporarily save SP */   \
	lmw	%r2, (FRAME_2+8)(%r1);		/* Restore r2-r31 */	    \
	ldreg	%r0, (FRAME_0+8)(%r1);		/* Restore r0 */	    \
	mfsprg1	%r1;				/* Fixup new SRR0 */	    \
	mtsrr0	%r1;							    \
	mfsprg2	%r1;				/* Fixup new SRR1 */	    \
	mtsrr1	%r1;							    \
	mfsprg3	%r1				/* Restore SP */


/*
 * INTR_SETUP assumes that the cpu registers are in the same state
 * as set up by the first level interrupt trampoline:
 *
 *	SPRG1: Original SRR0
 *	SPRG2: Original SRR1
 *	SPRG3: Original R1
 *	R1:    Spare
 */
#define	INTR_SETUP							    \
	GET_CPUINFO(%r1);						    \
	stmw	%r28, CI_TEMPSAVE(%r1);		/* Save r28 -> r31 */	    \
	mflr	%r29;				/* Save LR */		    \
	mfcr	%r30;				/* Save CR */		    \
	mfxer	%r31;				/* Save XER */		    \
	ldint	%r28, CI_INTRDEPTH(%r1);	/* Already on intstk? */    \
	addic.	%r28, %r28, 1;						    \
	stint	%r28, CI_INTRDEPTH(%r1);				    \
	ldptr	%r1, CI_INTSTK(%r1);		/* Get intstk */	    \
	beq	1f;				/* Yes, SP is good */	    \
	mfsprg3	%r1;				/* Otherwise, use old SP */ \
1:	stwu	%r1, -IFRAMELEN(%r1);					    \
	stw	%r0, IFRAME_R0(%r1);					    \
	mfsprg3	%r0;							    \
	stw	%r0, IFRAME_R1(%r1);					    \
	stw	%r3, IFRAME_R3(%r1);					    \
	stw	%r4, IFRAME_R4(%r1);					    \
	stw	%r5, IFRAME_R5(%r1);					    \
	stw	%r6, IFRAME_R6(%r1);					    \
	stw	%r7, IFRAME_R7(%r1);					    \
	stw	%r8, IFRAME_R8(%r1);					    \
	stw	%r9, IFRAME_R9(%r1);					    \
	stw	%r10, IFRAME_R10(%r1);					    \
	stw	%r11, IFRAME_R11(%r1);					    \
	stw	%r12, IFRAME_R12(%r1);					    \
	stw	%r28, IFRAME_INTR_DEPTH(%r1);				    \
	stw	%r29, IFRAME_LR(%r1);					    \
	stw	%r30, IFRAME_CR(%r1);					    \
	stw	%r31, IFRAME_XER(%r1);					    \
	GET_CPUINFO(%r3);						    \
	lmw	%r28, CI_TEMPSAVE(%r3);		/* Restore r28 -> r31 */    \
	mfctr	%r4;							    \
	mfsprg1	%r5;				/* Fetch saved SRR0 */	    \
	mfsprg2	%r6;				/* Fetch saved SRR1 */	    \
	stw	%r4, IFRAME_CTR(%r1);		/* Save CTR */		    \
	stw	%r5, IFRAME_SRR0(%r1);		/* Save SRR0 */		    \
	stw	%r6, IFRAME_SRR1(%r1);		/* Save SRR1 */		    \
	mfmsr	%r4;							    \
	ori	%r4, %r4, (PSL_ME|PSL_RI);				    \
	mtmsr	%r4;							    \
	isync


	.text

/*
 * exc_dtlbmiss - Data TLB Miss Handler
 */
	.globl	_C_LABEL(exc_dtlbmiss), _C_LABEL(exc_dtlbmiss_size)
_C_LABEL(exc_dtlbmiss):
	mtspr	SPR_MD_TW, %r1		/* Save r1 */
	mfcr	%r1
	streg	%r1, 0(0)		/* Save CR */
	streg	%r2, 4(0)		/* Save r2 */
	mfspr	%r1, SPR_M_TWB		/* Fetch L1 descriptor address */
	andi.	%r2, %r1, M_TWB_L1INDX	/* Keep L1 index */
	cmpwi	%r2, 0xc00		/* Kernel space miss? */
	blt+	1f			/* Jump if not */
	lis	%r1, (_C_LABEL(pmap_kernel_l1_pa)-PPC_KSEG0_BASE)@ha
	ldint	%r1, (_C_LABEL(pmap_kernel_l1_pa)-PPC_KSEG0_BASE)@l(%r1)
	add	%r1, %r1, %r2		/* Use kernel L1 */
1:	ldint	%r1, 0(%r1)		/* Fetch appropriate ptp entry */
	andi.	%r2, %r1, L1_VALID	/* Valid? */
	beq-	2f			/* Nope. */
	mtspr	SPR_MD_TWC, %r1		/* Save L2 base & L1 attributes */
	mfspr	%r1, SPR_MD_TWC		/* Fetch ptr to appropriate L2 pte */
	ldint	%r1, 0(%r1)		/* Fetch PTE */
	andi.	%r2, %r1, L2_VALID	/* Valid? */
	beq-	2f			/* Nope. */
	mtspr	SPR_MD_RPN, %r1		/* Write TLB entry */
	ldreg	%r2, 4(0)		/* Restore r2 */
	ldreg	%r1, 0(0)		/* Restore CR */
	mtcr	%r1
	mfspr	%r1, SPR_MD_TW		/* Restore r1 */
	rfi

	/* Have to take a full-blown trap */
2:	mfspr	%r1, SPR_MD_EPN		/* Save page number in DAR */
	mfdar   %r2
	rlwimi	%r2, %r1, 0, 0, 19
	mtdar   %r2

	ldreg	%r2, 4(0)		/* Restore r2 */
	ldreg	%r1, 0(0)		/* Restore CR */
	mtcr	%r1
	mfspr	%r1, SPR_MD_TW		/* Restore r1 */
	b	(EXC_DSI - (EXC_DTMISS8xx + (. - _C_LABEL(exc_dtlbmiss))))
_C_LABEL(exc_dtlbmiss_size) = .-_C_LABEL(exc_dtlbmiss)


/*
 * exc_itlbmiss - Instruction TLB Miss Handler
 */
	.globl	_C_LABEL(exc_itlbmiss), _C_LABEL(exc_itlbmiss_size)
_C_LABEL(exc_itlbmiss):
	mtspr	SPR_MD_TW, %r1		/* Save r1 */
	mfcr	%r1
	streg	%r1, 0(0)		/* Save CR */
	streg	%r2, 4(0)		/* Save r2 */
	mfsrr0	%r1			/* Fetch EA of missed instruction */
	mtspr	SPR_MD_EPN, %r1		/* Preload MD_EPN with miss address */
	mfspr	%r1, SPR_M_TWB		/* Fetch L1 ptp address */
	andi.	%r2, %r1, M_TWB_L1INDX	/* Keep L1 index */
	cmpwi	%r2, 0xc00		/* Kernel space miss? */
	blt+	1f			/* Jump if not */
	lis	%r1, (_C_LABEL(pmap_kernel_l1_pa)-PPC_KSEG0_BASE)@ha
	ldint	%r1, (_C_LABEL(pmap_kernel_l1_pa)-PPC_KSEG0_BASE)@l(%r1)
	add	%r1, %r1, %r2		/* Use kernel L1 */
1:	ldint	%r1, 0(%r1)		/* Fetch appropriate ptp entry */
	andi.	%r2, %r1, L1_VALID	/* Valid? */
	beq-	2f			/* Nope. */
	mtspr	SPR_MI_TWC, %r1		/* Save L1 attributes */
	mtspr	SPR_MD_TWC, %r1		/* Save L2 base */
	mfspr	%r1, SPR_MD_TWC		/* Fetch ptr to appropriate L2 pte */
	ldint	%r1, 0(%r1)		/* Fetch PTE */
	andi.	%r2, %r1, L2_VALID	/* Valid? */
	beq-	2f			/* Nope. */
	mtspr	SPR_MI_RPN, %r1		/* Write TLB entry */
	ldreg	%r2, 4(0)		/* Restore r2 */
	ldreg	%r1, 0(0)		/* Restore CR */
	mtcr	%r1
	mfspr	%r1, SPR_MD_TW		/* Restore r1 */
	rfi

	/* Have to take a full-blown trap */
2:	ldreg	%r2, 4(0)		/* Restore r2 */
	ldreg	%r1, 0(0)		/* Restore CR */
	mtcr	%r1
	mfspr	%r1, SPR_MD_TW		/* Restore r1 */
	b	(EXC_ISI - (EXC_ITMISS8xx + (. - _C_LABEL(exc_itlbmiss))))
_C_LABEL(exc_itlbmiss_size) = .-_C_LABEL(exc_itlbmiss)


/*
 * exc_trap_tramp - First level non-interrupt exception handler
 *
 * The following code is copied to all exception vectors except for
 * EXC_ITLBMISS, EXC_DTLBMISS, EXC_EXI, and EXC_DECR.
 *
 * Since exceptions are handled with the MMU switched off, the intention
 * here is to do the minimum work necessary to enable the MMU and bounce
 * back to real handler in kernel address space.
 *
 * When control is passed to the real exception handler, state is saved
 * in the following registers:
 *
 *	 SPRG1: Original SRR0
 *	 SPRG2: Original SRR1
 *	 SPRG3: Original R1
 *	    R1: scratch
 * R28/R29/R31: scratch
 *	   R30: Exception number
 * CI_TEMPSAVE: R28-R31, DAR, DSISR
 */
	.globl	_C_LABEL(exc_trap_tramp), _C_LABEL(exc_trap_tramp_size)
	.globl	_C_LABEL(exc_trap_tramp_handler_offset)
	.globl	_C_LABEL(exc_trap_tramp_number_offset)
_C_LABEL(exc_trap_tramp):
	mtsprg3	%r1			/* Save r1 */
	mfsrr0	%r1			/* Fetch SRR0 */
	mtsprg1	%r1			/* Save it */
	mfsrr1	%r1			/* Fetch SRR1 */
	mtsprg2	%r1			/* Save it */
	mfmsr	%r1			/* Fetch current MSR */
	ori	%r1, %r1, PSL_DR|PSL_IR	/* Enable translation */
	mtsrr1	%r1			/* Fixup MSR to restore on rfi */
	lis	%r1, (_C_LABEL(cpu_info)-PPC_KSEG0_BASE)@ha
	addi	%r1, %r1, (_C_LABEL(cpu_info)-PPC_KSEG0_BASE)@l
	stmw	%r28, CI_TEMPSAVE(%r1)	/* Save r28 -> r31 */
	mfdar	%r30			/* Save DAR */
	mfdsisr	%r31			/* Save DSISR */
	stmw	%r30, (CI_TEMPSAVE+16)(%r1) /* Copy to savearea */
_C_LABEL(exc_trap_tramp_handler_offset) = .-_C_LABEL(exc_trap_tramp)
	lis	%r28, _C_LABEL(exc_handler)@h
	ori	%r28, %r28, _C_LABEL(exc_handler)@l
	mtsrr0	%r28			/* Fixup insn. address to restore */
_C_LABEL(exc_trap_tramp_number_offset) = .-_C_LABEL(exc_trap_tramp)
	li	%r30, 0			/* Exception number: set in C code */
	isync
	rfi
_C_LABEL(exc_trap_tramp_size) = .-_C_LABEL(exc_trap_tramp)


/*
 * exc_handler - Second level non-interrupt exception handler
 *
 * Control is passed to this code from the first level exception handlers.
 * The MMU is enabled, and state has been saved as described above:
 *	 SPRG1: Original SRR0
 *	 SPRG2: Original SRR1
 *	 SPRG3: Original R1
 *	    R1: scratch
 * R28/R29/R31: scratch
 *	   R30: Exception number
 * CI_TEMPSAVE: R28-R31, DAR, DSISR
 */
ENTRY_NOPROFILE(exc_handler)
.Lreal_exc_handler:
	mflr	%r28			/* Save LR */
	mfcr	%r29			/* Save CR */
	mfsprg2	%r1			/* Fetch original MSR */
	mtcr	%r1			/* Copy to CR */
	mfsprg3	%r1			/* Restore original r1 */
	bf	17, .Lk_trap		/* Jump if came from supervisor mode */
.Lu_trap:
	GET_CPUINFO(%r1)
	ldptr	%r1, CI_CURPCB(%r1)	
	addi	%r1, %r1, USPACE	/* Top of kernel stack */
.Lk_trap:
	FRAME_SETUP(CI_TEMPSAVE)

	/* Exceptions are recoverable at this point */
	mfmsr	%r2
	ori	%r2, %r2, (PSL_ME|PSL_RI)@l
	mtmsr	%r2
	isync

.Ltrapagain:
	/* Enable interrupts, if they were previously enabled */
	mfmsr	%r2
	ldreg	%r3, (FRAME_SRR1+8)(%r1)
	andi.	%r3, %r3, PSL_EE@l	/* Retain previous PSL_EE state */
	or	%r2, %r2, %r3
	mtmsr	%r2

	addi	%r3, %r1, (2*SZREG)
	bl	_C_LABEL(trap)

	.globl	_C_LABEL(trapexit)
_C_LABEL(trapexit):
	mfmsr	%r2
	andi.	%r2, %r2, ~PSL_EE@l	/* Disable interrupts */
	mtmsr	%r2
	ldreg	%r3, (FRAME_SRR1+8)(%r1)
	andi.	%r3, %r3, PSL_PR	/* Came from user-mode? */
	beq	.Ltrapdone		/* Nope. Just return */
	GET_CPUINFO(%r3)
	ldint	%r3, CI_ASTPENDING(%r3)
	andi.	%r3, %r3, 0x01
	beq	.Ltrapdone		/* Return if no ASTs pending */
	li	%r3, EXC_AST
	streg	%r3, (FRAME_EXC+8)(%r1)
	b	.Ltrapagain
.Ltrapdone:
	FRAME_LEAVE
	rfi


/*
 * exc_sc_handler - Second level System Call handler
 *
 * Control is passed to this code from the first level exception handler.
 * The MMU is enabled, and state has been saved as described above:
 *	 SPRG1: Original SRR0
 *	 SPRG2: Original SRR1
 *	 SPRG3: Original R1
 *	    R1: scratch
 * R28/R29/R31: scratch
 *	   R30: Exception number
 * CI_TEMPSAVE: R28-R31, DAR, DSISR
 */
	.globl	_C_LABEL(sctrapexit)

ENTRY_NOPROFILE(exc_sc_handler)
	mflr	%r28			/* Save LR */
	mfcr	%r29			/* Save CR */
	GET_CPUINFO(%r1)
	ldptr	%r1, CI_CURPCB(%r1)	
	addi	%r1, %r1, USPACE	/* Top of kernel stack */
	FRAME_SETUP(CI_TEMPSAVE)

	/* Exceptions are recoverable at this point */
	mfmsr	%r2
	ori	%r2, %r2, (PSL_EE|PSL_ME|PSL_RI)@l
	mtmsr	%r2
	isync

	addi	%r3, %r1, (2*SZREG)
	/* Call the appropriate syscall handler: */
	GET_CPUINFO(%r4)
	ldptr	%r4, CI_CURLWP(%r4)
	ldptr	%r4, L_PROC(%r4)
	ldptr	%r4, P_MD_SYSCALL(%r4)
	mtctr	%r4
	bctrl

_C_LABEL(sctrapexit):
	mfmsr	%r3
	andi.	%r3, %r3, ~PSL_EE@l	/* Disable interrupts */
	mtmsr	%r3
	GET_CPUINFO(%r3)
	ldint	%r3, CI_ASTPENDING(%r3)
	andi.	%r3, %r3, 0x01
	beq	.Lscdone		/* Return if no ASTs pending */
	li	%r3, EXC_AST
	streg	%r3, (FRAME_EXC+8)(%r1)
	b	.Ltrapagain
.Lscdone:
	FRAME_LEAVE
	rfi


/*
 * exc_intr_tramp - First level interrupt exception handler
 *
 * The following code is copied to the EXC_EXI, and EXC_DECR vectors.
 *
 * Since exceptions are handled with the MMU switched off, the intention
 * here is to do the minimum work necessary to enable the MMU and bounce
 * back to real handler in kernel address space.
 *
 * When control is passed to the real exception handler, state is saved
 * in the following registers:
 *
 *	SPRG1: Original SRR0
 *	SPRG2: Original SRR1
 *	SPRG3: Original R1
 */
	.globl	_C_LABEL(exc_intr_tramp), _C_LABEL(exc_intr_tramp_size)
	.globl	_C_LABEL(exc_intr_tramp_handler_offset)
_C_LABEL(exc_intr_tramp):
	mtsprg3	%r1			/* Save r1 */
	mfsrr0	%r1			/* Fetch SRR0 */
	mtsprg1	%r1			/* Save it */
	mfsrr1	%r1			/* Fetch SRR1 */
	mtsprg2	%r1			/* Save it */
	mfmsr	%r1			/* Fetch current MSR */
	ori	%r1, %r1, PSL_DR|PSL_IR	/* Enable translation */
	mtsrr1	%r1			/* Fixup MSR to restore on rfi */
_C_LABEL(exc_intr_tramp_handler_offset) = .-_C_LABEL(exc_intr_tramp)
	lis	%r1, 0x0@h
	ori	%r1, %r1, 0x0@l
	mtsrr0	%r1			/* Fixup insn. address to restore */
	isync
	rfi
_C_LABEL(exc_intr_tramp_size) = .-_C_LABEL(exc_intr_tramp)


/*
 * extint_handler - Second level interrupt exception handler
 *
 * Control is passed to this code from the first level exception handler.
 * The MMU is enabled, and state has been saved in SPRG1-3 and R1 is
 * available as a scratch register.
 */
	.globl	_C_LABEL(extint_call)

ENTRY_NOPROFILE(extint_handler)
	INTR_SETUP
	mr	%r3, %r1		/* (struct intrframe *) */
_C_LABEL(extint_call):
	bl	_C_LABEL(extint_call)
	/* FALLTHROUGH */

.Lintr_exit:
	mfmsr	%r3
	andi.	%r3, %r3, ~(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsr	%r3
	isync
	ldreg	%r12, IFRAME_R12(%r1)
	ldreg	%r11, IFRAME_R11(%r1)
	ldreg	%r10, IFRAME_R10(%r1)
	ldreg	%r9, IFRAME_R9(%r1)
	ldreg	%r8, IFRAME_R8(%r1)
	ldreg	%r7, IFRAME_R7(%r1)
	ldreg	%r6, IFRAME_SRR1(%r1)
	ldreg	%r5, IFRAME_SRR0(%r1)
	ldreg	%r4, IFRAME_CTR(%r1)
	ldreg	%r3, IFRAME_XER(%r1)
	mtsprg1	%r5
	ldreg	%r5, IFRAME_R1(%r1)
	mtsprg2	%r6
	mtctr	%r4
	mtxer	%r3
	mtsprg3	%r5
	GET_CPUINFO(%r3)
	ldint	%r4, CI_INTRDEPTH(%r3)
	addi	%r4, %r4, -1
	stint	%r4, CI_INTRDEPTH(%r3)
	mtcr	%r6
	bf	17, 1f			/* Jump if came from supervisor mode */

	/* Returing to user mode */
	ldint	%r4, CI_ASTPENDING(%r3)
	andi.	%r4, %r4, 1
	bne-	2f			/* AST pending */

1:	ldreg	%r5, IFRAME_CR(%r1)
	ldreg	%r6, IFRAME_LR(%r1)
	mtcr	%r5
	mtlr	%r6
	ldreg	%r6, IFRAME_R6(%r1)
	ldreg	%r5, IFRAME_R5(%r1)
	ldreg	%r4, IFRAME_R4(%r1)
	ldreg	%r3, IFRAME_R3(%r1)
	ldreg	%r0, IFRAME_R0(%r1)
	mfsprg1	%r1				/* Fixup new SRR0 */
	mtsrr0	%r1
	mfsprg2	%r1				/* Fixup new SRR1 */
	mtsrr1	%r1
	mfsprg3	%r1				/* Restore SP */
	rfi

	/* Restore state such that we can re-enter via .Lreal_exc_handler */
2:	ldreg	%r5, IFRAME_CR(%r1)
	ldreg	%r6, IFRAME_LR(%r1)
	mtcr	%r5
	mtlr	%r6
	ldreg	%r6, IFRAME_R6(%r1)
	ldreg	%r5, IFRAME_R5(%r1)
	ldreg	%r4, IFRAME_R4(%r1)
	ldreg	%r3, IFRAME_R3(%r1)
	ldreg	%r0, IFRAME_R0(%r1)
	GET_CPUINFO(%r1)
	stmw	%r28, CI_TEMPSAVE(%r1)	/* Save r28 -> r31 */
	li	%r30, 0			/* Zero DAR */
	li	%r31, 0			/* Zero DSISR */
	stmw	%r30, (CI_TEMPSAVE+16)(%r1) /* Copy to savearea */
	li	%r30, EXC_AST
	b	.Lreal_exc_handler


/*
 * decrint_handler - Second level decrementer interrupt exception handler
 *
 * Control is passed to this code from the first level exception handler.
 * The MMU is enabled, and state has been saved in SPRG1-3 and R1 is
 * available as a scratch register.
 */
ENTRY_NOPROFILE(decrint_handler)
	INTR_SETUP
	addi	%r3, %r1, (2*SZREG)	/* intr frame -> clock frame */
	bl	_C_LABEL(decr_intr)
	b	.Lintr_exit


#ifdef DDB
/*
 * ddb_trap - Deliberate entry to DDB
 */
ENTRY_NOPROFILE(ddb_trap)
	mfmsr	%r3
	andi.	%r4, %r3, ~(PSL_ME|PSL_RI|PSL_EE)@l	/* Disable interrupts */
	mtmsr	%r4
	sync
	mtsprg3	%r1			/* Save r1 */
	mtsprg2	%r3			/* Save MSR */
	mflr	%r1
	mtsprg1	%r1			/* Save LR */
	GET_CPUINFO(%r1)
	stmw	%r28, CI_TEMPSAVE(%r1)	/* Save r28 -> r31 */
	li	%r30, 0
	li	%r31, 0
	stmw	%r30, (CI_TEMPSAVE+16)(%r1)
	li	%r30, EXC_TRC
	b	_C_LABEL(exc_ddb_handler)
	nop


/*
 * exc_ddb_handler - Second level TRACE exception handler
 *
 * Control is passed to this code from the first level exception handler.
 * The MMU is enabled, and state has been saved as described above.
 *	 SPRG1: Original SRR0
 *	 SPRG2: Original SRR1
 *	 SPRG3: Original R1
 *	    R1: scratch
 * R28/R29/R31: scratch
 *	   R30: Exception number
 * CI_TEMPSAVE: R28-R31, DAR, DSISR
 */
ENTRY_NOPROFILE(exc_ddb_handler)
	mflr	%r28			/* Save LR */
	mfcr	%r29			/* Save CR */
	mfsprg2	%r31
	mtcr	%r31
	bt	17, .Lu_trap		/* Jump if came from user mode */
	mtcr	%r29			/* Restore CR for now */
	GET_CPUINFO(%r1)
	streg	%r30, (CI_DDBSAVE+24)(%r1)	/* Save exception number */
	lmw	%r28, CI_TEMPSAVE(%r1)	/* Restore r28 -> r31 */
	stmw	%r28, CI_DDBSAVE(%r1)	/* Save r28 -> r31 in rightful place */
	lmw	%r30, (CI_TEMPSAVE+16)(%r1)
	stmw	%r30, (CI_DDBSAVE+16)(%r1)
	ldreg	%r30, (CI_DDBSAVE+24)(%r1)
	mflr	%r28			/* Save LR */
	mfcr	%r29			/* Save CR */
	mfsprg3	%r1			/* Restore original SP */
	FRAME_SETUP(CI_DDBSAVE)

	/* Exceptions are recoverable at this point */
	mfmsr	%r4
	ori	%r4, %r4, (PSL_ME|PSL_RI)@l
	mtmsr	%r4
	isync

	addi	%r3, %r1, (2*SZREG)
	bl	_C_LABEL(ddb_trap_glue)

	mfmsr	%r4
	andi.	%r4, %r4, ~(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsr	%r4
	isync
	or.	%r3, %r3, %r3
	bne	.Lddb_done
	FRAME_LEAVE
	GET_CPUINFO(%r1)
	stmw	%r28, CI_TEMPSAVE(%r1)	/* Save r28 -> r31 */
	li	%r30, 0			/* Zero DAR */
	li	%r31, 0			/* Zero DSISR */
	stmw	%r30, (CI_TEMPSAVE+16)(%r1) /* Copy to savearea */
	li	%r30, EXC_TRC		/* TRACE exception number */
	b	.Lreal_exc_handler
.Lddb_done:
	FRAME_LEAVE
	rfi


/*
 * Read the TLB into the supplied buffer
 *
 * struct mpc8xx_tlb {
 *	uint32_t mi_cam;
 *	uint32_t mi_ram0;
 *	uint32_t mi_ram1;
 *	uint32_t md_cam;
 *	uint32_t md_ram0;
 *	uint32_t md_ram1;
 * };
 *
 * void ddb_read_tlb(paddr_t)
 */
ENTRY(ddb_read_tlb)
	mfmsr	%r4
	andi.	%r6, %r4, ~(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsr	%r6
	lis	%r5, 1f-PPC_KSEG0_BASE@ha
	addi	%r5, %r5, 1f-PPC_KSEG0_BASE@l
	mtsrr0	%r5
	andi.	%r6, %r6, ~(PSL_DR|PSL_IR)@l
	mtsrr1	%r6
	isync
	rfi

	/* Read the TLB with the MMU off */
1:	li	%r5, 0x0
	mfspr	%r7, SPR_MI_CTR		/* Save MI_CTR */
	mfspr	%r8, SPR_MD_CTR		/* Save MI_CTR */
2:	mfspr	%r6, SPR_MI_CTR		/* Fetch MI_CTR */
	rlwinm	%r6, %r6, 0, 24, 18	/* Zap ITLB_INDX */
	or	%r6, %r6, %r5		/* Include required TLB index */
	mtspr	SPR_MI_CTR, %r6
	mtspr	SPR_MI_CAM, %r6
	mfspr	%r6, SPR_MI_CAM		/* Read CAM enty at specified index */
	streg	%r6, 0x00(%r3)
	mfspr	%r6, SPR_MI_RAM0	/* Ditto RAM0 */
	streg	%r6, 0x04(%r3)
	mfspr	%r6, SPR_MI_RAM1	/* And RAM1 */
	streg	%r6, 0x08(%r3)

	mfspr	%r6, SPR_MD_CTR		/* Fetch MD_CTR */
	rlwinm	%r6, %r6, 0, 24, 18	/* Zap DTLB_INDX */
	or	%r6, %r6, %r5		/* Include required TLB index */
	mtspr	SPR_MD_CTR, %r6
	mtspr	SPR_MD_CAM, %r6
	mfspr	%r6, SPR_MD_CAM		/* Read CAM entry at specified index */
	streg	%r6, 0x0c(%r3)
	mfspr	%r6, SPR_MD_RAM0	/* Ditto RAM0 */
	streg	%r6, 0x10(%r3)
	mfspr	%r6, SPR_MD_RAM1	/* And RAM1 */
	streg	%r6, 0x14(%r3)

	addi	%r3, %r3, 0x18		/* Advance array pointer */
	addi	%r5, %r5, 0x0100	/* Next TLB index */
	cmpwi	%r5, 0x2000		/* Done all 32 entries? */
	bne	2b
	mtspr	SPR_MI_CTR, %r7		/* Restore MI_CTR */
	mtspr	SPR_MD_CTR, %r8		/* Restore MI_CTR */
	mflr	%r5			/* Set up to return to caller */
	mtsrr0	%r5
	mtsrr1	%r4			/* r4 held original MSR */
	isync
	rfi
#endif /* DDB */
